# Barton Willis, Copyright 2016
# This work is licensed under a Creative Commons Attribution 4.0 International License.
# Julia code for the evaluation of the hypergeometric functions 2F1 and 3F2.


#module Hypergeometric
#export f32_polynomial_case, f32, f32_mp

#Extend eps to rationals and integers--the machine epsilon is zero for these numbers

# In R, use library(hypergeo); options(digits=10); hypergeo(-100, 4.0,0.1,0.3)
import Base.eps
eps(x::Int) = zero(x)
eps(x::Rational) = zero(x)
eps(x::BigInt) = zero(x) 

function f32_polynomial_case(a,b,c,d,e,x)
    # among a,b, and c find the largest negative integer.
    (a,b,c) = sort([a,b,c], lt = (x,y) -> isa(x,Int) && x <=0 && x > y)
    println("a=$a, b=$b, c=$c")
    if !isa(a,Int) || a > 0
        throw(ErrorException("Nonpolynomial case given to f32_polynomial_case."))  
    end
    s = zero(x)
    z = one(x)
    k = 0
    while (k <= -a)
        s = s + z
        z = z * x * (a+k)*(b+k)*(c+k)/((d+k)*(e+k)*(1+k))
        k = k + 1
    end 
    s
end 

function f32 (a,b,c,d,e,x,N::Int64=typemax(Int64))
    #a, b, c, d, e, x = promote(a, b, c, d, e, x)
    (c,a,b) = sort([a,b,c], by = real)
    
    z = if (x == conj(x)) x/2 else abs2(x)/(conj(x)-x) - (x/(conj(x)-x)) * (1 - abs(x-1)) end
    w = z/(z-1)
    cnd = max(abs2(z/(z-1)), abs2((z-x)/(z-1)))
    if cnd >= 1 throw(DomainError()) end
    
    f0 = one(x) #3F2(0,b,c; d,e; x/z)
    q = one(x)  #pochhammer(a,0) * w^0 / 0!
    s = one(x)  #partial sum
    
    f1 = one(x)-(b*c*x)/(d*e*z) #3F2(-1,b,c; d,e; x/z)
    q = a * w #pochhammer(a,1)*w^1/ 1!
    s += q * f1
    
    f2 = -(2*b*c*x)/(d*e*z)-((-b-1)*b*c*(c+1)*x^2)/(d*(d+1)*e*(e+1)*z^2)+1 #3F2(-2,b,c; d,e; x/z)
    q *= (a+1) * w / 2 #pochhammer(a,2) * w^2 / 2!
    s0 = s
    s += q * f2
    
    kk = zero(N)
    er = zero(x)
    ep = max(eps(real(x)), eps(imag(x)))
    
    const A0 = a + 2 #hoist some constants by hand; symmetric in d <-> e.
    const A1 = z-x
    const A2 = c+b+3
    const A3 = e+d+4 
    const A4 = 2*(e+d) + 11
    const A5 = 4*(d+e) + e*d + 10
    const A6 = b + 2
    const A7 = c + 2
    const A8 = d + 2
    const A9 = e + 2
    const xtype = typeof(x)   
    while (kk < N) && (abs(s-s0) > abs(s) * ep)
        k = convert(xtype,kk) #for efficiency
        q *= (k+A0)*w/(k+3) 
        P0 = (k+1)*(k+2)*A1
        P1 = (k+2)*(2*k+A2)*x-(k+2)*(3*k+A3)*z
        P2 = (k*(3*k+A4)+A5)*z-(k+A6)*(k+A7)*x   
        f3 = (P0*f0 + P1*f1 + P2*f2)/((k+A8)*(k+A9)*z) #could do sorted add...
        (f0,f1,f2) = (f1,f2,f3)
        kk += 1
        s0 = s
        s += q * f2
        er += abs(s)
    end
    q = (1-z)^a #recycle q--ok
    s = s / q
    er = ep * er / q
    s,er,kk
end
        
# multiprecsion version of f32
function f32_mp (a, b, c, d, e, x, digits::Int64 = 15)
    bfprec = get_bigfloat_precision()  #used to restore value of initial bigfloat precision.
    set_bigfloat_precision(int(digits/log10(2.0))) #set the value of bigfloat precsion to something reasonable
    (aa, bb, cc, dd, ee, xx) = promote(a, b, c, d, e, x) 
    return_type  = typeof(xx)
    #println("aa = $aa bb = $bb cc = $cc xx = $xx")
    (f, er, k) = f32(aa,bb,cc,dd,ee, xx)

    while abs(er) * 10.0^digits > max(1, abs(er) + abs(f)) #declare do-over
        #heuristic
        n = get_bigfloat_precision() +
        if (abs(er)/max(1,abs(f)) > 0.1)
            2*get_bigfloat_precision()
        else
            max(int(-log2(max(1, abs(er) + abs(f)) / (10.0^digits * abs(er)))), 32) #magic number 32
        end
        info("setting bigfloat precision to $n")
        set_bigfloat_precision(n)
        (aa,bb,cc,xx) = map(x -> convert(Complex{BigFloat},x),(a,b,c,x))
        f, er, k = f32(aa,bb,cc,dd, ee, xx)
    end
    f = convert(return_type,f)
    set_bigfloat_precision(bfprec) #restore value of bigfloat precision
    f #return
end
#end #of hypergeometric module
